package wood.keith.opentools.wizards.jsptags;

import java.io.IOException;
import java.lang.reflect.Modifier;
import java.text.DateFormat;
import java.text.MessageFormat;
import java.util.Date;
import java.util.ResourceBundle;

import com.borland.jbuilder.jot.JotClassSource;
import com.borland.jbuilder.jot.JotCodeBlock;
import com.borland.jbuilder.jot.JotComment;
import com.borland.jbuilder.jot.JotIf;
import com.borland.jbuilder.jot.JotMethodSource;
import com.borland.jbuilder.jot.JotPackages;
import com.borland.jbuilder.jot.JotSourceFile;
import com.borland.jbuilder.jot.JotVariableDeclaration;
import com.borland.jbuilder.node.JBProject;
import com.borland.primetime.node.FileNode;

/**
 * Generate the JSP tag extra information class.
 *
 * @author   Keith Wood
 * @version  3.0  10 February 2004 - added JSP 2.0 support
 */
public class JSPTagExtraGenerator {

  private static final ResourceBundle res = Res.getResource();
  // Meaningful names for code generation
  private static final boolean AFTER = false;
  private static final boolean BEFORE = true;

  /**
   * This is the "guts" of the wizard. This is where all the work
   * is done to create the source file for the tag class.
   *
   * @param  project      the JBuilder project
   * @param  file         the file node to be written to
   * @param  packageName  the name of the new class' package
   * @param  className    the name of the new class
   * @param  jspVersion   the version of JSP in use
   */
  public void writeSource(JBProject project, FileNode file,
      String packageName, String className, String jspVersion) {
    // Access the new file
    JotPackages jpackages = project.getJotPackages();
    JotSourceFile jsource = jpackages.getSourceFile(file.getUrl());
    // Add the package statement and imports to the JotSourceFile
    if (packageName.length() > 0) {
      jsource.setPackage(packageName);
    }
    jsource.addImport("java.util.*");
    jsource.addImport("javax.servlet.jsp.tagext.*");

    // Add the class declaration
    JotClassSource jclass = createClass(jsource, className);
    // Add method overrides
    addGetVariableInfo(jclass);
    if (jspVersion.equals(JSPTagProperties.JSP_20)) {
      addValidate(jclass);
    }
    else {
      addIsValid(jclass);
    }

    // Save the changes
    jpackages.commit(jsource);

    // Write the changes out
    try {
      file.save();
    }
    catch (IOException ex) {
      ex.printStackTrace();
    }
    // Free up resources used
    jpackages.release(jsource);
  }

  /**
   * Add the class declaration.
   *
   * @param  jsource    the source file being generated
   * @param  className  the name of the new class
   * @return  the class being generated
   */
  private JotClassSource createClass(JotSourceFile jsource, String className) {
    JotClassSource jclass = jsource.addClass(null, AFTER, className, false);
    jsource.addComment(jclass, BEFORE, JotComment.DOC, 
        MessageFormat.format(res.getString("JSPExtraClassHeader"), 
        new Object[] {className.substring(0, className.length() - 9)}) +
        "\n \n@author   JSP Tag Wizard\n@version  1.0  " +
        DateFormat.getDateInstance(DateFormat.LONG).format(new Date()));
    jclass.setModifiers(Modifier.PUBLIC);
    jclass.setSuperclass("TagExtraInfo");
    return jclass;
  }

  /**
   * Add TagExtraInfo method getVariableInfo override.
   *
   * @param  jclass  the class being generated
   */
  private void addGetVariableInfo(JotClassSource jclass) {
    JotMethodSource jmethod = jclass.addMethod(null, AFTER, "VariableInfo[]", 
        "getVariableInfo");
    JotComment jcomment = jclass.addComment(jmethod, BEFORE, JotComment.DOC,
        res.getString("GetVariableInfoMethod") + "\n \n@param  data  " +
        res.getString("TagDataParam") + "\n@return  " +
        res.getString("GetVariableInfoReturn"));
    jclass.addBlankLine(jcomment, BEFORE);
    jmethod.setModifiers(Modifier.PUBLIC);
    jmethod.addParameter(null, AFTER, "TagData", "data");
    JotCodeBlock jcode = jmethod.getCodeBlock();
    JotVariableDeclaration jvar =
        jcode.addVariableDeclaration(null, AFTER, "ArrayList", "vars");
    jvar.setInitializer("new ArrayList()");
    jcode.addComment(null, AFTER, JotComment.DOC, 
        " @todo " + res.getString("GetVariableInfoTodo") + " ");
    jcode.addComment(null, AFTER, JotComment.LINE, 
        "vars.add(new VariableInfo(varName, className, declare, scope));");
    jcode.addReturnStatement(null, AFTER, 
        "(VariableInfo[])vars.toArray(new VariableInfo[vars.size()])");
  }

  /**
   * Add TagExtraInfo method validate override.
   *
   * @param  jclass  the class being generated
   */
  private void addValidate(JotClassSource jclass) {
    JotMethodSource jmethod = jclass.addMethod(null, AFTER, 
        "ValidationMessage[]", "validate");
    JotComment jcomment = jclass.addComment(jmethod, BEFORE, JotComment.DOC,
        res.getString("ValidateMethod") + "\n \n@param  data  " +
        res.getString("TagDataParam") + "\n@return  " +
        res.getString("ValidateReturn"));
    jclass.addBlankLine(jcomment, BEFORE);
    jmethod.setModifiers(Modifier.PUBLIC);
    jmethod.addParameter(null, AFTER, "TagData", "data");
    JotCodeBlock jcode = jmethod.getCodeBlock();
    JotVariableDeclaration jvar =
        jcode.addVariableDeclaration(null, AFTER, "ArrayList", "errors");
    jvar.setInitializer("new ArrayList()");
    jcode.addComment(null, AFTER, JotComment.DOC, 
        " @todo " + res.getString("ValidateTodo") + " ");
    jcode.addComment(null, AFTER, JotComment.LINE, "if (invalid) {");
    jcode.addComment(null, AFTER, JotComment.LINE,
        "  errors.add(new ValidationMessage(id, message));");
    jcode.addComment(null, AFTER, JotComment.LINE, "}");
    jcode.addReturnStatement(null, AFTER, "(ValidationMessage[])" +
        "errors.toArray(new ValidationMessage[errors.size()])");
  }

  /**
   * Add TagExtraInfo method isValid override.
   *
   * @param  jclass  the class being generated
   */
  private void addIsValid(JotClassSource jclass) {
    JotMethodSource jmethod =
        jclass.addMethod(null, AFTER, "boolean", "isValid");
    JotComment jcomment = jclass.addComment(jmethod, BEFORE, JotComment.DOC,
        res.getString("IsValidMethod") + "\n \n@param  data  " +
        res.getString("TagDataParam") + "\n@return  " + 
        res.getString("IsValidReturn"));
    jclass.addBlankLine(jcomment, BEFORE);
    jmethod.setModifiers(Modifier.PUBLIC);
    jmethod.addParameter(null, AFTER, "TagData", "data");
    JotCodeBlock jcode = jmethod.getCodeBlock();
    jcode.addComment(null, AFTER, JotComment.DOC, 
        " @todo " + res.getString("IsValidTodo") + " ");
    JotIf jif = jcode.addIfStatement(null, AFTER, "validates");
    jif.getCodeBlock().addReturnStatement(null, AFTER, "true");
    jcode.addReturnStatement(null, AFTER, "false");
  }
}
