package wood.keith.opentools.wizards.jsptags;

import java.io.IOException;
import java.lang.reflect.Modifier;
import java.text.DateFormat;
import java.text.MessageFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.ResourceBundle;
import java.util.SortedMap;

import com.borland.jbuilder.jot.JotClassSource;
import com.borland.jbuilder.jot.JotCodeBlock;
import com.borland.jbuilder.jot.JotComment;
import com.borland.jbuilder.jot.JotFieldDeclaration;
import com.borland.jbuilder.jot.JotIf;
import com.borland.jbuilder.jot.JotMethodSource;
import com.borland.jbuilder.jot.JotPackages;
import com.borland.jbuilder.jot.JotSourceFile;
import com.borland.jbuilder.jot.JotTry;
import com.borland.jbuilder.jot.JotVariableDeclaration;
import com.borland.jbuilder.jot.JotWhile;
import com.borland.jbuilder.node.JBProject;
import com.borland.primetime.node.FileNode;

/**
 * Generate the JSP tag class.
 *
 * @author   Keith Wood
 * @version  1.0  9 November 2000
 * @version  2.0  6 February 2002 - added JSP 1.2 support
 * @version  3.0  10 February 2004 - added JSP 2.0 support
 */
public class JSPTagGenerator {

  private static final ResourceBundle res = Res.getResource();
  // Meaningful names for code generation
  private static final boolean AFTER = false;
  private static final boolean BEFORE = true;
  
  private static final String[] OUTER_TAG = new String[] {"OuterTag"};

  /**
   * This is the "guts" of the wizard. This is where all the work
   * is done to create the source file for the tag class.
   *
   * @param  project            the JBuilder project
   * @param  file               the file node to be written to
   * @param  packageName        the name of the new class' package
   * @param  className          the name of the new class
   * @param  jspVersion         the version of JSP in use
   * @param  isBodyAltered      true if body content is altered, false if not
   * @param  isRepeated         true if body is repeated, false if not
   * @param  isEndUsed          true if end tag generated, false if not
   * @param  isNested           true if tag expects to be nested, false if not
   * @param  isTryCatchFinally  true if implementing TryCatchFinally,
   *                            false if not
   * @param  isSimpleTag        true if a JSP 2.0 simple tag, false if not
   * @param  isDynamicAttrs     true if uses dynamic attributes, false if not
   * @param  props              details about the tag attributes
   */
  public void writeSource(JBProject project, FileNode file,
      String packageName, String className, String jspVersion,
      boolean isBodyAltered, boolean isRepeated, boolean isEndUsed,
      boolean isNested, boolean isTryCatchFinally, boolean isSimpleTag,
      boolean isDynamicAttrs, SortedMap props) {
    boolean hasProperties = false;
    boolean hasDateAttribute = false;

    Iterator names = props.keySet().iterator();
    while (names.hasNext()) {
      hasProperties = true;
      if (props.get(names.next()).equals("Date")) {
        hasDateAttribute = true;
        break;
      }
    }
    names = null;

    // Access the new file
    JotPackages jpackages = project.getJotPackages();
    JotSourceFile jsource = jpackages.getSourceFile(file.getUrl());
    // Add the package statement and imports to the JotSourceFile
    if (packageName.length() > 0) {
      jsource.setPackage(packageName);
    }
    jsource.addImport("java.io.*");
    if (hasDateAttribute) {
      jsource.addImport("java.util.*");
    }
    jsource.addImport("javax.servlet.jsp.*");
    jsource.addImport("javax.servlet.jsp.tagext.*");

    // Add the class declaration
    JotClassSource jclass = createClass(jsource, className, jspVersion, 
        isEndUsed, isNested, isRepeated, isBodyAltered, isTryCatchFinally, 
        isSimpleTag, isDynamicAttrs);
    if (hasProperties || isNested) {
      // Add internal fields
      addInternalFields(jclass, props, isNested);
    }
    // Add attribute setters
    addAttributeSetters(jclass, props);
    if (isDynamicAttrs) {
      // Add DynamicAttributes method setDynamicAttribute
      addSetDynamicAttributeMethod(jclass);
    }
    if (isSimpleTag) {
      // Add SimpleTag method doTag override
      addDoTag(jclass, isNested, isRepeated, isBodyAltered, hasProperties);
    }
    else {
      // Add Tag method doStartTag override
      addDoStartTag(jclass, jspVersion, isEndUsed, isNested, isRepeated,
          isBodyAltered, hasProperties);
      if (isBodyAltered || isRepeated) {
        // Add BodyTag method doInitBody override
        addDoInitBody(jclass);
        // Add BodyTag method doAfterBody override
        addDoAfterBody(jclass, jspVersion, isRepeated);
      }
      if (isEndUsed) {
        // Add Tag method doEndTag override
        addDoEndTag(jclass);
      }
    }
    if (isTryCatchFinally) {
      // Add TryCatchFinally methods doCatch and doFinally
      addTryCatchFinallyMethods(jclass);
    }

    // Save the changes
    jpackages.commit(jsource);

    // Write the changes out
    try {
      file.save();
    }
    catch (IOException ex) {
      ex.printStackTrace();
    }
    // Free up resources used
    jpackages.release(jsource);
  }

  /**
   * Add the class declaration.
   *
   * @param  jsource            the source file being generated
   * @param  className          the name of the new class
   * @param  jspVersion         the version of JSP in use
   * @param  isEndUsed          true if content is written after the body,
   *                            false if none
   * @param  isNested           true if the tag looks for an enclosing tag,
   *                            false if standalone
   * @param  isRepeated         true if the tag iterates over its body,
   *                            false if it processes it just once (if at all)
   * @param  isBodyAltered      true if the tag changes the body content,
   *                            false if it leaves it unchanged
   * @param  isTryCatchFinally  true if the tag deals with its own errors during
   *                            processing, false if it does not
   * @param  isSimpleTag        true if tag is a JSP 2.0 simple tag,
   *                            false if not
   * @param  isDynamicAttrs     true if tag accepts unspecified attributes,
   *                            false if not
   * @return  the class being generated
   */
  private JotClassSource createClass(JotSourceFile jsource, String className,
      String jspVersion, boolean isEndUsed, boolean isNested,
      boolean isRepeated, boolean isBodyAltered, boolean isTryCatchFinally,
      boolean isSimpleTag, boolean isDynamicAttrs) {
    JotClassSource jclass = jsource.addClass(null, AFTER, className, false);
    StringBuffer comment = new StringBuffer();
    comment.append(isEndUsed ? res.getString("TagIsEndUsed") : "").
        append(isNested ? res.getString("TagIsNested") : "").
        append(isRepeated ? res.getString("TagIsRepeated") : "").
        append(isBodyAltered ? res.getString("TagIsBodyAltered") : "").
        append(isTryCatchFinally ? res.getString("TagIsTryCatchFinally") : "").
        append(isSimpleTag ? res.getString("TagIsSimpleTag") : "").
        append(isDynamicAttrs ? res.getString("TagIsDynamicAttrs") : "");
    jsource.addComment(jclass, BEFORE, JotComment.DOC, 
        MessageFormat.format(res.getString("JSPTagClassHeader"), 
        new Object[] {jspVersion}) + (comment.length() == 0 ? "" :
        " " + res.getString("JSPTagClassHeaderCharacteristics") + 
        comment.toString()) + 
        ".\n \n@author   JSP Tag Wizard\n@version  1.0  " +
        DateFormat.getDateInstance(DateFormat.LONG).format(new Date()));
    jclass.setModifiers(Modifier.PUBLIC);
    jclass.setSuperclass(isSimpleTag ? "SimpleTagSupport" : 
        (isBodyAltered || isRepeated ? "BodyTagSupport" : "TagSupport"));
    if (isDynamicAttrs) {
      jclass.addInterface(null, AFTER, "DynamicAttributes");
    }
    if (isTryCatchFinally) {
      jclass.addInterface(null, AFTER, "TryCatchFinally");
    }
    return jclass;
  }

  /**
   * Add internal fields for the tag class.
   *
   * @param  jclass    the class being generated
   * @param  props     the names and types of the properties for this tag
   * @param  isNested  true if the tag looks for an enclosing tag,
   *                   false if standalone
   */
  private void addInternalFields(JotClassSource jclass, SortedMap props,
      boolean isNested) {
    JotComment jcomment = jclass.addComment(null, AFTER, JotComment.LINE, 
        " " + res.getString("InternalFields"));
    jclass.addBlankLine(jcomment, BEFORE);
    if (props.size() > 0) {
      jclass.addComment(null, AFTER, JotComment.DOC,
          " @todo " + res.getString("InternalFieldsTodo") + " ");
    }
    Iterator names = props.keySet().iterator();
    while (names.hasNext()) {
      String name = (String)names.next();
      JotFieldDeclaration jfield =
          jclass.addField(null, AFTER, (String)props.get(name), "_" + name);
      jfield.setModifiers(Modifier.PRIVATE);
    }
    if (isNested) {
      jclass.addComment(null, AFTER, JotComment.DOC, 
          " @todo " + MessageFormat.format(res.getString("OuterTagTodo"), 
          OUTER_TAG) + " ");
      JotFieldDeclaration jfield =
          jclass.addField(null, AFTER, "OuterTag", "_enclosingTag");
      jfield.setModifiers(Modifier.PRIVATE);
      jfield.setInitializer("null");
    }
  }

  /**
   * Add attribute setters for the tag.
   *
   * @param  jclass  the class being generated
   * @param  props   the names and types of the properties for this tag
   */
  private void addAttributeSetters(JotClassSource jclass, SortedMap props) {
    Iterator names = props.keySet().iterator();
    while (names.hasNext()) {
      String name = (String)names.next();
      JotMethodSource jmethod = jclass.addMethod(null, AFTER, "void",
          "set" + name.substring(0, 1).toUpperCase() + name.substring(1));
      JotComment jcomment = jclass.addComment(jmethod, BEFORE, JotComment.DOC,
          MessageFormat.format(res.getString("AttributeSetterMethod"),
          new Object[] {name}) + "\n \n@param  value  " +
          res.getString("ValueParam"));
      jclass.addBlankLine(jcomment, BEFORE);
      jmethod.setModifiers(Modifier.PUBLIC);
      jmethod.addParameter(null, AFTER, (String)props.get(name), "value");
      jmethod.getCodeBlock().addAssignment(null, AFTER, "_" + name, "value");
    }
  }

  /**
   * Add DynamicAttributes method setDynamicAttribute.
   *
   * @param  jclass  the class being generated
   */
  private void addSetDynamicAttributeMethod(JotClassSource jclass) {
    JotMethodSource jmethod = 
        jclass.addMethod(null, AFTER, "void", "setDynamicAttribute");
    JotComment jcomment = jclass.addComment(jmethod, BEFORE, JotComment.DOC,
        res.getString("SetDynamicAttributeMethod") + "\n \n@param  uri        " +
        res.getString("UriParam") + "\n@param  localName  " +
        res.getString("LocalNameParam") + "\n@param  value      " +
        res.getString("ValueParam") + "\n@throws  JspException  " +
        res.getString("SetDynamicAttributeJspException"));
    jclass.addBlankLine(jcomment, BEFORE);
    jmethod.setModifiers(Modifier.PUBLIC);
    jmethod.addParameter(null, AFTER, "String", "uri");
    jmethod.addParameter(null, AFTER, "String", "localName");
    jmethod.addParameter(null, AFTER, "Object", "value");
    jmethod.addThrowSpecifier(null, AFTER, "JspException");
    JotCodeBlock jcode = jmethod.getCodeBlock();
    jcode.addComment(null, AFTER, JotComment.DOC,
        " @todo " + res.getString("SetDynamicAttributeTodo") + " ");
    jcode.addStatement(null, AFTER, "throw new JspException(\"" + 
        MessageFormat.format(res.getString("SetDynamicAttributeThrow"), 
        new Object[] {"\" + localName + \"", "\" + uri + \""}) + "\")");
  }

  /**
   * Add SimpleTag method doTag override.
   *
   * @param  jclass         the class being generated
   * @param  isNested       true if the tag looks for an enclosing tag,
   *                        false if standalone
   * @param  isRepeated     true if the tag iterates over its body,
   *                        false if it processes it just once (if at all)
   * @param  isBodyAltered  true if the tag changes the body content,
   *                        false if it leaves it unchanged
   * @param  hasProperties  true if the tag has properties, false otherwise
   */
  private void addDoTag(JotClassSource jclass, boolean isNested, 
      boolean isRepeated, boolean isBodyAltered, boolean hasProperties) {
    JotMethodSource jmethod = jclass.addMethod(null, AFTER, "void", "doTag");
    JotComment jcomment = jclass.addComment(jmethod, BEFORE, JotComment.DOC,
        res.getString("DoTagMethod") + 
        (hasProperties ? "\n" + res.getString("DoTagHasProperties") : "") +
        "\n \n@throws  IOException        " + res.getString("DoTagIOException") +
        "\n@throws  JSPException       " + res.getString("DoTagJspException") +
        "\n@throws  SkipPageException  " + 
        res.getString("DoTagSkipPageException"));
    jclass.addBlankLine(jcomment, BEFORE);
    jmethod.setModifiers(Modifier.PUBLIC);
    jmethod.addThrowSpecifier(null, AFTER, "IOException");
    jmethod.addThrowSpecifier(null, AFTER, "JspException");
    JotCodeBlock jcode = jmethod.getCodeBlock();
    if (isNested) {
      jcode.addComment(null, AFTER, JotComment.LINE,
          res.getString("DoTagFindEnclosing"));
      jcode.addComment(null, AFTER, JotComment.DOC,
          " @todo " + MessageFormat.format(res.getString("OuterTagTodo"), 
          OUTER_TAG) + " ");
      jcode.addAssignment(null, AFTER, "_enclosingTag",
          "(OuterTag)findAncestorWithClass(this, OuterTag.class)");
      JotIf jif = jcode.addIfStatement(null, AFTER, "_enclosingTag == null");
      jif.getThen().getCodeBlock().addStatement(null, AFTER,
          "throw new JspException(\"" + 
          MessageFormat.format(res.getString("DoTagThrowNestingError"), 
          OUTER_TAG) + "\")");
    }
    JotVariableDeclaration jvar =
        jcode.addVariableDeclaration(null, AFTER, "JspWriter", "out");
    jvar.setInitializer("getJspContext().getOut()");

    if (isRepeated || isBodyAltered) {
      jcode.addComment(null, AFTER, JotComment.DOC,
          " @todo " + res.getString("TagOpeningOutputTodo") + " ");
      jcode.addStatement(null, AFTER, "out.print(\"" + 
          res.getString("TagOpeningOutputValue") + "\")");
      jvar = jcode.addVariableDeclaration(null, AFTER, "JspFragment", "body");
      jvar.setInitializer("getJspBody()");
      if (isRepeated) {
        jcode.addComment(null, AFTER, JotComment.DOC,
            " @todo " + res.getString("RepeatInitTodo") + " ");
        jcode.addComment(null, AFTER, JotComment.DOC,
            " @todo " + res.getString("RepeatTestTodo") + " ");
        JotWhile jwhile = jcode.addWhileStatement(null, AFTER, "repeatBody");
        jcode = jwhile.getCodeBlock();
      }
      jcode.addComment(null, AFTER, JotComment.DOC,
          " @todo " + res.getString("DoTagAccessBodyTodo") + " ");
      jcode.addMethodCall(null, AFTER, "body.invoke", "out");
      jcode = jmethod.getCodeBlock();
      jcode.addComment(null, AFTER, JotComment.DOC,
          " @todo " + res.getString("TagClosingOutputTodo") + " ");
      jcode.addStatement(null, AFTER, "out.print(\"" +
          res.getString("TagClosingOutputValue") + "\")");
    }
    else {
      jcode.addComment(null, AFTER, JotComment.DOC,
          " @todo " + res.getString("TagOutputTodo") + " ");
      jcode.addStatement(null, AFTER, "out.print(\"" +
          res.getString("TagOutputValue") + "\")");
    }
  }

  /**
   * Add Tag method doStartTag override.
   *
   * @param  jclass         the class being generated
   * @param  jspVersion     the version of JSP in use
   * @param  isEndUsed      true if content is written after the body,
   *                        false if none
   * @param  isNested       true if the tag looks for an enclosing tag,
   *                        false if standalone
   * @param  isRepeated     true if the tag iterates over its body,
   *                        false if it processes it just once (if at all)
   * @param  isBodyAltered  true if the tag changes the body content,
   *                        false if it leaves it unchanged
   * @param  hasProperties  true if the tag has properties, false otherwise
   */
  private void addDoStartTag(JotClassSource jclass, String jspVersion,
      boolean isEndUsed, boolean isNested, boolean isRepeated,
      boolean isBodyAltered, boolean hasProperties) {
    boolean isJSP11 = jspVersion.equals(JSPTagProperties.JSP_11);
    JotMethodSource jmethod = jclass.addMethod(null, AFTER, 
        "int", "doStartTag");
    JotComment jcomment = jclass.addComment(jmethod, BEFORE, JotComment.DOC,
        res.getString("DoStartTagMethod") + "\n \n" +
        (hasProperties ? res.getString("DoStartTagHasProperties") : "") + 
        "\n \n@return  " +
        (!isBodyAltered && !isRepeated ? "" : MessageFormat.format(
        res.getString("DoStartTagReturnBody"), new Object[] 
        {(isJSP11 ? "EVAL_BODY_TAG" : "EVAL_BODY_BUFFERED"), 
        (isJSP11 ? res.getString("DoStartTagReturnIterates") : "")}) + 
        "\n         ") +
        MessageFormat.format(res.getString("DoStartTagReturn"), new Object[] 
        {"EVAL_BODY_INCLUDE", "SKIP_BODY"}) + "\n@throws  JspException  " +
        res.getString("DoStartTagJspException"));
    jclass.addBlankLine(jcomment, BEFORE);
    jmethod.setModifiers(Modifier.PUBLIC);
    jmethod.addThrowSpecifier(null, AFTER, "JspException");
    JotTry jtry = jmethod.getCodeBlock().
        addTryStatement(null, AFTER, "IOException", "ex");
    JotCodeBlock jcode = jtry.getCodeBlock();
    JotIf jif = null;
    if (isNested) {
      jcode.addComment(null, AFTER, JotComment.LINE,
          " " + res.getString("DoTagFindEnclosing"));
      jcode.addComment(null, AFTER, JotComment.DOC,
          " @todo " + MessageFormat.format(res.getString("OuterTagTodo"), 
          OUTER_TAG) + " ");
      jcode.addAssignment(null, AFTER, "_enclosingTag",
          "(OuterTag)findAncestorWithClass(this, OuterTag.class)");
      jif = jcode.addIfStatement(null, AFTER, "_enclosingTag == null");
      jif.getThen().getCodeBlock().addStatement(null, AFTER,
          "throw new JspException(\"" + 
          MessageFormat.format(res.getString("DoTagThrowNestingError"), 
          OUTER_TAG) + "\")");
    }
    JotVariableDeclaration jvar =
        jcode.addVariableDeclaration(null, AFTER, "JspWriter", "out");
    jvar.setInitializer("pageContext.getOut()");
    jcode.addComment(null, AFTER, JotComment.DOC,
        " @todo " + res.getString("TagOpeningOutputTodo") + " ");
    jcode.addStatement(null, AFTER, "out.print(\"" +
        (isEndUsed ? res.getString("TagOpeningOutputValue") : 
        res.getString("TagOutputValue")) + "\")");
    jtry.getCatches()[0].getCodeBlock().
        addStatement(null, AFTER, "ex.printStackTrace()");
    jtry.getCatches()[0].getCodeBlock().
        addStatement(null, AFTER, "throw new JspException(ex.getMessage())");

    if (isRepeated) {
      jcode.addComment(null, AFTER, JotComment.DOC,
          " @todo " + res.getString("RepeatInitTodo") + " ");
      jcode.addComment(null, AFTER, JotComment.DOC,
          " @todo " + res.getString("RepeatTestTodo") + " ");
      jif = jcode.addIfStatement(null, AFTER, "repeatBody", true);
      jcode = jif.getThen().getCodeBlock();
      jcode.addComment(null, AFTER, JotComment.DOC,
          " @todo " + res.getString("DoStartTagFirstIterationTodo") + " ");
      jcode.addReturnStatement(null, AFTER, 
          (isJSP11 ? "EVAL_BODY_TAG" : "EVAL_BODY_BUFFERED"));
      jif.getElse().getCodeBlock().
          addReturnStatement(null, AFTER, "SKIP_BODY");
    }
    else {
      jcode.addReturnStatement(null, AFTER, (isBodyAltered ?
          (isJSP11 ? "EVAL_BODY_TAG" : "EVAL_BODY_BUFFERED") :
          (isEndUsed ? "EVAL_BODY_INCLUDE" : "SKIP_BODY")));
    }
  }

  /**
   * Add Tag method doInitBody override.
   *
   * @param  jclass the class being generated
   */
  private void addDoInitBody(JotClassSource jclass) {
    JotMethodSource jmethod =
        jclass.addMethod(null, AFTER, "void", "doInitBody");
    JotComment jcomment = jclass.addComment(jmethod, BEFORE, JotComment.DOC,
        res.getString("DoInitBodyMethod") + "\n \n@throws  JspException  " +
        res.getString("DoInitBodyJspException"));
    jclass.addBlankLine(jcomment, BEFORE);
    jmethod.setModifiers(Modifier.PUBLIC);
    jmethod.addThrowSpecifier(null, AFTER, "JspException");
  }

  /**
   * Add Tag method doAfterBody override.
   *
   * @param  jclass      the class being generated
   * @param  jspVersion  the version of JSP in use
   * @param  isRepeated  true if the tag iterates over its body,
   *                     false if it processes it just once (if at all)
   */
  private void addDoAfterBody(JotClassSource jclass, String jspVersion,
      boolean isRepeated) {
    boolean isJSP11 = jspVersion.equals(JSPTagProperties.JSP_11);
    JotMethodSource jmethod =
        jclass.addMethod(null, AFTER, "int", "doAfterBody");
    JotComment jcomment = jclass.addComment(jmethod, BEFORE, JotComment.DOC,
        MessageFormat.format(res.getString("DoAfterBodyMethod"), 
        new Object[] {"SKIP_BODY", 
        (isJSP11 ? "EVAL_BODY_TAG" : "EVAL_BODY_INCLUDE/EVAL_BODY_BUFFERED"),
        (isJSP11 ? "EVAL_BODY_TAG" : "EVAL_BODY_AGAIN")}) + "\n \n@return  " +
        MessageFormat.format(res.getString("DoAfterBodyReturn"), new Object[] 
        {(isJSP11 ? "EVAL_BODY_TAG" : "EVAL_BODY_AGAIN"), "SKIP_BODY"}) + 
        "\n@throws  JspException  " + res.getString("DoAfterBodyJspException"));
    jclass.addBlankLine(jcomment, BEFORE);
    jmethod.setModifiers(Modifier.PUBLIC);
    jmethod.addThrowSpecifier(null, AFTER, "JspException");
    JotTry jtry = jmethod.getCodeBlock().
        addTryStatement(null, AFTER, "IOException", "ex");
    JotCodeBlock jcode = jtry.getCodeBlock();
    JotVariableDeclaration jvar =
        jcode.addVariableDeclaration(null, AFTER, "BodyContent", "body");
    jvar.setInitializer("getBodyContent()");
    jvar = jcode.addVariableDeclaration(null, AFTER, "JspWriter", "out");
    jvar.setInitializer("body.getEnclosingWriter()");
    jcode.addComment(null, AFTER, JotComment.DOC,
        " @todo " + res.getString("DoAfterBodyChangeBodyTodo") + " ");
    jcode.addStatement(null, AFTER, "out.print(body.getString())");
    jcode.addStatement(null, AFTER, "body.clearBody()");
    if (isRepeated) {
      jcode.addComment(null, AFTER, JotComment.DOC,
          " @todo " + res.getString("RepeatTestTodo") + " ");
      JotIf jif = jcode.addIfStatement(null, AFTER, "repeatBody", true);
      jcode = jif.getThen().getCodeBlock();
      jcode.addComment(null, AFTER, JotComment.DOC,
          " @todo " + res.getString("RepeatNextIterationTodo") + " ");
      jcode.addReturnStatement(null, AFTER,
          (isJSP11 ? "EVAL_BODY_TAG" : "EVAL_BODY_AGAIN"));
      jcode = jif.getElse().getCodeBlock();
    }
    jcode.addReturnStatement(null, AFTER, "SKIP_BODY");
    jtry.getCatches()[0].getCodeBlock().
        addStatement(null, AFTER, "ex.printStackTrace()");
    jtry.getCatches()[0].getCodeBlock().
        addStatement(null, AFTER, "throw new JspException(ex.getMessage())");
  }

  /**
   * Add Tag method doEndTag override.
   *
   * @param  jclass  the class being generated
   */
  private void addDoEndTag(JotClassSource jclass) {
    JotMethodSource jmethod = jclass.addMethod(null, AFTER, "int", "doEndTag");
    JotComment jcomment = jclass.addComment(jmethod, BEFORE, JotComment.DOC,
        res.getString("DoEndTagMethod") + "\n \n@return  " +
        MessageFormat.format(res.getString("DoEndTagReturn"), 
        new Object[] {"EVAL_PAGE", "SKIP_PAGE"}) + 
        "\n@throws  JspException  " + res.getString("DoEndTagJspException"));
    jclass.addBlankLine(jcomment, BEFORE);
    jmethod.setModifiers(Modifier.PUBLIC);
    jmethod.addThrowSpecifier(null, AFTER, "JspException");
    JotTry jtry = jmethod.getCodeBlock().
        addTryStatement(null, AFTER, "IOException", "ex");
    JotCodeBlock jcode = jtry.getCodeBlock();
    JotVariableDeclaration jvar =
        jcode.addVariableDeclaration(null, AFTER, "JspWriter", "out");
    jvar.setInitializer("pageContext.getOut()");
    jcode.addComment(null, AFTER, JotComment.DOC,
        " @todo " + res.getString("TagClosingOutputTodo") + " ");
    jcode.addStatement(null, AFTER, "out.print(\"" +
        res.getString("TagClosingOutputValue") + "\")");
    jcode.addReturnStatement(null, AFTER, "EVAL_PAGE");
    jtry.getCatches()[0].getCodeBlock().
        addStatement(null, AFTER, "ex.printStackTrace()");
    jtry.getCatches()[0].getCodeBlock().
        addStatement(null, AFTER, "throw new JspException(ex.getMessage())");
  }

  /**
   * Add TryCatchFinally methods doCatch and doFinally.
   *
   * @param  jclass  the class being generated
   */
  private void addTryCatchFinallyMethods(JotClassSource jclass) {
    JotMethodSource jmethod = jclass.addMethod(null, AFTER, "void", "doCatch");
    JotComment jcomment = jclass.addComment(jmethod, BEFORE, JotComment.DOC,
        res.getString("DoCatchMethod") + "\n \n@param  th  " +
        res.getString("ThrowableParam"));
    jclass.addBlankLine(jcomment, BEFORE);
    jmethod.setModifiers(Modifier.PUBLIC);
    jmethod.addParameter(null, AFTER, "Throwable", "th");
    JotCodeBlock jcode = jmethod.getCodeBlock();
    jcode.addComment(null, AFTER, JotComment.DOC,
        " @todo " + res.getString("DoCatchTodo") + " ");

    jmethod = jclass.addMethod(null, AFTER, "void", "doFinally");
    jcomment = jclass.addComment(jmethod, BEFORE, JotComment.DOC,
        " " + res.getString("DoFinallyMethod") + " ");
    jclass.addBlankLine(jcomment, BEFORE);
    jmethod.setModifiers(Modifier.PUBLIC);
    jcode = jmethod.getCodeBlock();
    jcode.addComment(null, AFTER, JotComment.DOC,
        " @todo " + res.getString("DoFinallyTodo") + " ");
  }
}
