package wood.keith.opentools.wizards.jsptags;

import java.io.File;
import java.util.ResourceBundle;
import javax.swing.ImageIcon;
import javax.swing.JOptionPane;

import com.borland.jbuilder.node.JBProject;
import com.borland.jbuilder.paths.ProjectPathSet;
import com.borland.jbuilder.personality.WebAppPersonality;
import com.borland.primetime.PrimeTime;
import com.borland.primetime.help.HelpManager;
import com.borland.primetime.help.ZipHelpTopic;
import com.borland.primetime.ide.Browser;
import com.borland.primetime.node.FileNode;
import com.borland.primetime.personality.Personality;
import com.borland.primetime.util.VetoException;
import com.borland.primetime.vfs.Url;
import com.borland.primetime.wizard.BasicWizard;
import com.borland.primetime.wizard.Wizard;
import com.borland.primetime.wizard.WizardAction;
import com.borland.primetime.wizard.WizardHost;
import com.borland.primetime.wizard.WizardManager;
import com.borland.primetime.wizard.WizardPage;

/**
 * This class defines the "JSP Tag" wizard at the highest level.
 * It sets the dialog pages to be displayed and generates the
 * JSP tag class as defined by user input.
 *
 * @author   Keith Wood
 * @version  1.0  9 November 2000
 * @version  2.0  6 February 2002 - JSP 1.2
 * @version  3.0  10 February 2004 - JSP 2.0
 */
public class JSPTagWizard extends BasicWizard {

  private static final ResourceBundle res = Res.getResource();
  private static final String VERSION = "3.0";

  public static final String TITLE = res.getString("JSPWizardTitle");

  // Internal variables
  private JSPTagWizardPage _tagPage;

  /**
   * Register the "JSP Tag" tool.
   *
   * Provides the needed OpenTools interface required to register the
   * WizardAction which defines and creates this wizard.
   *
   * @param  majorVersion  the major version of the current OpenTools API
   * @param  minorVersion  the minor version of the current OpenTools API
   */
  public static void initOpenTool(byte majorVersion, byte minorVersion) {
    if (majorVersion != PrimeTime.CURRENT_MAJOR_VERSION) {
      return;
    }
    WizardManager.registerWizardAction(WIZARD_JSPTag);
    if (PrimeTime.isVerbose()) {
      System.out.println("Loaded JSP Tag wizard v" + VERSION);
      System.out.println("Written by Keith Wood (kbwood@iprimus.com.au)");
    }
  }

  /**
   * Definition of "JSP Tag" WizardAction.
   *
   * This is an OpenTools registered action defining where the wizard will
   * appear in the IDE and a factory for creating each instance of this wizard.
   */
  public static final WizardAction WIZARD_JSPTag = new WizardAction(
      res.getString("JSPWizardName"), res.getString("JSPWizardChar").charAt(0),
      res.getString("JSPWizardTip"),
      new ImageIcon(ClassLoader.getSystemResource(
      "wood/keith/opentools/wizards/jsptags/jsptag16.gif")),
      new ImageIcon(ClassLoader.getSystemResource(
      "wood/keith/opentools/wizards/jsptags/jsptag32.gif")), true,
      res.getString("ObjectGalleryPage")) {

    // Override method so wizard is disabled if no active project.
    public void update(Object source) {
      Browser browser = Browser.findBrowser(source);
      setEnabled((browser != null) && (browser.getActiveProject() != null));
    }

    protected Wizard createWizard() {
      return new JSPTagWizard();
    }

    /**
     * JBuilder 10+ - only display for the Web app personality.
     *
     * @return  the list of personalities to which this action applies
     */
    public Personality[] getPersonalities() {
      return new Personality[] {WebAppPersonality.webAppPersonality};
    }
  };

  /**
   * Wizard startup.
   *
   * "JSP Tag" wizard is to become visible,
   * create and order the wizard pages to be displayed.
   *
   * @param  host  the WizardHost that owns this wizard instance.
   * @return  the initial WizardPage to show.
   */
  public WizardPage invokeWizard(WizardHost host) {
    setWizardTitle(TITLE);
    _tagPage = new JSPTagWizardPage(
        host.getBrowser().getProjectView().getActiveProject());
    addWizardPage(_tagPage);
    return super.invokeWizard(host);
  }

  /**
   * Assemble file Url.
   *
   * From the given parameters generates a fully qualified location
   * where the "JSP Tag" file will be written.
   * This is returned in the form of a Url.
   *
   * @param  project      the JBProject in which the node is created.
   * @param  packageName  the name of the package in which file is created.
   * @param  fileName     the name of the file to be created.
   * @return  the Url which was created.
   */
  protected static Url makeFileUrl(JBProject project, String packageName,
      String fileName) {
    try {
      ProjectPathSet paths = project.getPaths();
      Url[] sourcePaths = paths.getSourcePath();
      String projectPath = "";

      if (sourcePaths.length > 0) {
        projectPath = sourcePaths[0].getFile();
      }

      String dirName = projectPath;
      if (packageName != null && packageName.length() > 0) {
        dirName = dirName + '/' + packageName.replace('.', '/');
      }
      String filePath = dirName + '/' + fileName;
      File newFile = new File(filePath);
      return new Url(newFile);
    }
    catch (Exception ex) {
      ex.printStackTrace();
      return null;
    }
  }

  /**
   * Creates/replaces file node.
   *
   * From the given parameters create/replace a project file node for the
   * "JSP Tag" file to be written.
   *
   * @param  project      the JBProject in which the node is created.
   * @param  packageName  the name of the package in which file is created.
   * @param  fileName     the name of the file to be created.
   * @return  the FileNode which was created.
   */
  protected static FileNode createNode(JBProject project,
      String packageName, String fileName) {
    Url filePath = makeFileUrl(project, packageName, fileName);
    // getNode() will create a node if does not exist
    FileNode newNode = project.getNode(filePath);
    newNode.setParent(project);
    return newNode;
  }

  /**
   * Perform code generation.
   *
   * Produces the "JSP Tag" file as defined by user input on the
   * wizard pages. It is added as a node to the currently active project
   * and replaces any previously generated file of the same name.
   *
   * @throws  VetoException  if unable to finish
   */
  protected void finish() throws VetoException {
    Browser browser = wizardHost.getBrowser();
    JBProject project = (JBProject)browser.getProjectView().getActiveProject();
    // Save selected JSP version for next time
    JSPTagProperties.JSP_VERSION.setValue(_tagPage.getJSPVersion());
    FileNode javaNode = null;
    try {
      if (_tagPage.isExtraInfo()) {
        // Create the extra info source file
        javaNode = createNode(project, _tagPage.getPackageName(),
            _tagPage.getClassName() + "ExtraInfo.java");
        new JSPTagExtraGenerator().writeSource(project, javaNode,
            _tagPage.getPackageName(), _tagPage.getClassName() + "ExtraInfo",
            _tagPage.getJSPVersion());
      }
      // Create the tag source file
      javaNode = createNode(project, _tagPage.getPackageName(),
          _tagPage.getClassName() + ".java");
      new JSPTagGenerator().writeSource(project, javaNode,
          _tagPage.getPackageName(), _tagPage.getClassName(),
          _tagPage.getJSPVersion(), _tagPage.isBodyAltered(),
          _tagPage.isRepeated(), _tagPage.isEndUsed(), _tagPage.isNested(),
          _tagPage.isTryCatchFinally(), _tagPage.isSimpleTag(),
          _tagPage.isDynamicAttributes(), _tagPage.getProps());
    }
    catch (Exception ex) {
      ex.printStackTrace();
      JOptionPane.showMessageDialog(wizardHost.getBrowser(),
          res.getString("GenerationError") + "\n" + ex.getClass().getName() +
          "\n" + ex.getMessage(), wizardTitle, JOptionPane.ERROR_MESSAGE);
    }
    try {
      // Open the generated/updated file in the Content Pane.
      browser.setActiveNode(javaNode, true);
    }
    catch (Exception ex) {
      // Ignore
    }
  }

  /**
   * Wizard termination.
   *
   * The Wizard has been completed (finished or cancelled) and any resource
   * cleanup to help the garbage collector can occur here.
   */
  public void wizardCompleted() {
  }

  /**
   * Provide help for the wizard.
   *
   * @param  page  the current page in the wizard
   * @param  host  the wizard host
   */
  public void help(WizardPage page, WizardHost host) {
    HelpManager.showHelp(new ZipHelpTopic(null, ClassLoader.getSystemResource(
        getClass().getName().replace('.', '/') + ".html").toString()),
        page.getPageComponent(host));
  }
}
