package wood.keith.opentools.wizards.jsptags;

import java.util.ArrayList;

/**
 * Generate the contents of the tag library descriptor file.
 *
 * @author   Keith Wood
 * @version  1.0
 * @version  3.0  10 February 2004 - JSP 2.0
 */
public class TaglibDescriptorGenerator extends TaglibDescriptorElement {

	/** Common header for taglib documents. */
  private static final String COMMON_XML_START =
		  "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
	  	"<!-- Generated by Taglib Descriptor Wizard -->\n";

  /** Header for JSP 1.1 taglib documents. */
	private static final String TAGLIB_11_XML =
      COMMON_XML_START +
  		"<!DOCTYPE taglib\n" +
  		"  PUBLIC \"-//Sun Microsystems, Inc.//DTD JSP Tag Library 1.1//EN\"\n" +
	  	"  \"http://java.sun.com/j2ee/dtds/web-jsptaglibrary_1_1.dtd\">\n" +
		  "<taglib>\n";

  /** Header for JSP 1.2 taglib documents. */
  private static final String TAGLIB_12_XML =
  		COMMON_XML_START +
      "<!DOCTYPE taglib\n" +
		  "  PUBLIC \"-//Sun Microsystems, Inc.//DTD JSP Tag Library 1.2//EN\"\n" +
  		"  \"http://java.sun.com/dtd/web-jsptaglibrary_1_2.dtd\">\n" +
	  	"<taglib>\n";

  /** Header for JSP 2.0 taglib documents. */
  private static final String TAGLIB_20_XML =
      COMMON_XML_START +
      "<taglib version=\"2.0\" xmlns=\"http://java.sun.com/xml/ns/j2ee\"\n" +
      "    xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n" +
      "    xsi:schemaLocation=\"http://java.sun.com/xml/ns/j2ee " +
      "http://java.sun.com/xml/ns/j2ee/web-jsptaglibrary_2_0.xsd\">\n";

  /** Common footer for taglib documents. */
  private static final String COMMON_XML_END =
      "</taglib>\n";

	// Internal variables
	private String _shortName;
	private String _uri;
	private TaglibDescriptorValidator _validator;
	private ArrayList _listeners;
	private ArrayList _tags;
  private ArrayList _tagFiles;
  private ArrayList _functions;

	/**
	 * Initialisation.
	 *
	 * @param  shortName      the short name for the tag library
	 * @param  uri            the identifying URI for the tag library
	 * @param  displayName    the display text for the tag library
	 * @param  smallIconName  the URL for a small icon for the tag library
	 * @param  largeIconName  the URL for a large icon for the tag library
	 * @param  description    an extended description of the tag library
	 * @param  validator      the validator details for the tag library
	 * @param  listeners      the list of classes that listen for tag library
	 *                        events, each entry is a TaglibDescriptorListener
	 * @param  tags           information about the tags within the library,
	 *                        each entry is a TaglibDescriptorTag
   * @param  tagFiles       the list of tag files for the tag library,
   *                        each entry is a TaglibDescriptorTagFile
   * @param  functions      the list of functions for the tag library,
   *                        each entry is a TaglibDescriptorFunction
	 */
	public TaglibDescriptorGenerator(
      String shortName, String uri, String displayName, 
      String smallIconName, String largeIconName, String description, 
      TaglibDescriptorValidator validator, ArrayList listeners, 
      ArrayList tags, ArrayList tagFiles, ArrayList functions) {
		_shortName = shortName;
		_uri = uri;
		setDisplayName(displayName);
		setSmallIcon(smallIconName);
		setLargeIcon(largeIconName);
		setDescription(description);
		_validator = validator;
		_listeners = listeners;
		_tags = tags;
    _tagFiles = tagFiles;
    _functions = functions;
	}

	/**
	 * Perform code generation.
	 * Produces the "Taglib Descriptor" file as defined by user input on the
	 * wizard pages.
	 *
   * @param  jspVersion  the version of JSP in use
	 * @return  the content for the tag library descriptor file
	 */
	public String toXml(String jspVersion) {
    if (jspVersion.equals(JSPTagProperties.JSP_20)) {
      return toXml20(jspVersion);
    }
    else if (jspVersion.equals(JSPTagProperties.JSP_12)) {
      return toXml12(jspVersion);
    }
    else {
      return toXml11(jspVersion);
    }
	}

  /**
   * Generate XML document for JSP 2.0.
   *
   * @param  jspVersion  the version of JSP in use
   * @return  the XML elements for the header
   */
  private String toXml20(String jspVersion) {
    StringBuffer buffer = new StringBuffer(TAGLIB_20_XML);
    addDescriptionGroup(buffer, "  ");
    addXmlElement(buffer, "tlib-version", "1.0", "  ");
    addXmlElement(buffer, "short-name", _shortName, "  ");
    addXmlElement(buffer, "uri", _uri, "  ");
    addValidatorToXml(buffer, jspVersion);
    addXmlElements(buffer, jspVersion, _listeners);
    addXmlElements(buffer, jspVersion, _tags);
    addXmlElements(buffer, jspVersion, _tagFiles);
    addXmlElements(buffer, jspVersion, _functions);
    buffer.append(COMMON_XML_END);
    return buffer.toString();
  }

  /**
   * Generate XML document for JSP 1.2.
   *
   * @param  jspVersion  the version of JSP in use
   * @return  the XML elements for the header
   */
  private String toXml12(String jspVersion) {
    StringBuffer buffer = new StringBuffer(TAGLIB_12_XML);
    addXmlElement(buffer, "tlib-version", "1.0", "  ");
    addXmlElement(buffer, "jsp-version", "1.2", "  ");
    addXmlElement(buffer, "short-name", _shortName, "  ");
    addXmlElement(buffer, "uri", _uri, "  ");
    addXmlElement(buffer, "display-name", getDisplayName(), "  ");
    addXmlElement(buffer, "small-icon", getSmallIcon(), "  ");
    addXmlElement(buffer, "large-icon", getLargeIcon(), "  ");
    addXmlElement(buffer, "description", getDescription(), "  ");
    addValidatorToXml(buffer, jspVersion);
    addXmlElements(buffer, jspVersion, _listeners);
    addXmlElements(buffer, jspVersion, _tags);
    buffer.append(COMMON_XML_END);
    return buffer.toString();
  }

  /**
   * Generate XML document for JSP 1.1.
   *
   * @param  jspVersion  the version of JSP in use
   * @return  the XML elements for the header
   */
  private String toXml11(String jspVersion) {
    StringBuffer buffer = new StringBuffer(TAGLIB_11_XML);
    addXmlElement(buffer, "tlibversion", "1.0", "  ");
    addXmlElement(buffer, "jspversion", "1.1", "  ");
    addXmlElement(buffer, "shortname", _shortName, "  ");
    addXmlElement(buffer, "uri", _uri, "  ");
    addXmlElement(buffer, "info", getDescription(), "  ");
    addXmlElements(buffer, jspVersion, _tags);
    buffer.append(COMMON_XML_END);
    return buffer.toString();
  }

	/**
	 * Generate XML elements for any validator for the tag library.
	 *
   * @param  buffer      the string buffer to add to
   * @param  jspVersion  the version of JSP in use
	 */
	private void addValidatorToXml(StringBuffer buffer, String jspVersion) {
    if (_validator != null) {
      buffer.append(_validator.toXml(jspVersion));
    }
	}
}
