package wood.keith.opentools.wizards.jsptags;

import java.io.File;
import java.io.FileFilter;
import java.util.ArrayList;
import java.util.Arrays;

import com.borland.jbuilder.jot.JotClass;
import com.borland.jbuilder.jot.JotClassSource;
import com.borland.jbuilder.jot.JotComment;
import com.borland.jbuilder.jot.JotFile;
import com.borland.jbuilder.jot.JotMethod;
import com.borland.jbuilder.jot.JotMethodSource;
import com.borland.jbuilder.jot.JotPackages;
import com.borland.jbuilder.jot.JotSourceFile;
import com.borland.jbuilder.jot.JotType;
import com.borland.jbuilder.node.JBProject;
import com.borland.jbuilder.node.JavaFileNode;
import com.borland.jbuilder.node.PackageNode;
import com.borland.primetime.node.Node;

/**
 * Scan the project for candidate JSP tag classes and extract their details.
 *
 * @author   Keith Wood
 * @version  1.0  9 November 2000
 * @version  2.0  6 February 2002 - JSP 1.2
 * @version  2.1  16 June 2002
 * @version  3.0  10 February 2004 - JSP 2.0
 */
public class TaglibDescriptorScanner {

  /** Tag base interfaces. */
  private static final String[] TAG_INTERFACES =
      {"javax.servlet.jsp.tagext.IterationTag", 
      "javax.servlet.jsp.tagext.SimpleTag", 
      "javax.servlet.jsp.tagext.Tag"};

  /** Tag interfaces that reference a body. */
  private static final String[] BODY_INTERFACES =
      {"javax.servlet.jsp.tagext.BodyTag",
      "javax.servlet.jsp.tagext.IterationTag"};

  /** Dynamic attributes interface. */
  private static final String[] DYNAMIC_INTERFACES =
      {"javax.servlet.jsp.tagext.DynamicAttributes"};

  /** Listener base interfaces. */
  private static final String[] LISTENER_INTERFACES =
      {"javax.servlet.http.HttpSessionActivationListener",
      "javax.servlet.http.HttpSessionAttributeListener",
      "javax.servlet.http.HttpSessionBindingListener",
      "javax.servlet.http.HttpSessionListener",
      "javax.servlet.ServletContextAttributeListener",
      "javax.servlet.ServletContextListener",
      "javax.servlet.ServletRequestAttributeListener",
      "javax.servlet.ServletRequestListener"};

	/** Standard set... methods in Tag/SimpleTag classes - to be ignored. */
	private static final String[] STD_METHODS =
  		{"setBodyContent", "setDynamicAttribute", "setId", "setJspBody", 
      "setJspContext", "setPageContext", "setParent", "setValue"};
  
  private static final boolean BEFORE = true;
  
  /**
   * Create a new scanner and perform the scan on a project.
   * 
   * @param  project  the project to scan
   */
  public TaglibDescriptorScanner(JBProject project) {
    scanProjectForTags(project);
    scanProjectForTagFiles(project);
  }

	// Internal variables
	private ArrayList _tags = new ArrayList();
	private ArrayList _validators = new ArrayList();
	private ArrayList _listeners = new ArrayList();
	private ArrayList _tagExtraInfos = new ArrayList();
  private ArrayList _tagFiles = new ArrayList();

	public ArrayList getTags() { return _tags; }
	public ArrayList getValidators() { return _validators; }
	public ArrayList getListeners() { return _listeners; }
  public ArrayList getTagExtraInfos() { return _tagExtraInfos; }
	public ArrayList getTagFiles() { return _tagFiles; }

  /**
   * Perform scan for JSP tag files within this project.
   *
   * @param  project  the project to scan
   */
  private void scanProjectForTagFiles(JBProject project) {
    scanDirectoryForTagFiles(new File(
        project.getUrl().getFileObject().getParentFile(), "WEB-INF/tags"));
    scanDirectoryForTagFiles(new File(
        project.getUrl().getFileObject().getParentFile(), "META-INF/tags"));
  }
    
  /**
   * Scan a directory for tag file candidates (*.tag, *.tagx).
   * Save any found for later use.
   * 
   * @param  dir  the directory to scan
   */
  private void scanDirectoryForTagFiles(File dir) {
    File[] files = dir.listFiles(new FileFilter() {
      public boolean accept(File pathname) {
        return (pathname.getAbsolutePath().endsWith(".tag") ||
            pathname.getAbsolutePath().endsWith(".tagx"));
      }
    });
    if (files == null) {
      return;
    }
    for (int index = 0; index < files.length; index++) {
      // Capture tag files
      String path = files[index].getAbsolutePath();
      // Get path relative to project directory
      path = path.substring(files[index].getParentFile().getParentFile().
          getParentFile().getAbsolutePath().length());
      _tagFiles.add(new TaglibDescriptorTagFile(files[index].getName(), path));
    }
  }
  
	/**
	 * Perform scan.
	 * Invokes the scan of the java source in this project for taglib candidates.
	 *
   * @param  project  the project to scan
	 */
	private void scanProjectForTags(JBProject project) {
		_tags.clear();
		_validators.clear();
		_listeners.clear();
		_tagExtraInfos.clear();
    _tagFiles.clear();
		scanNodesForTags(project, project.getChildren());
	}

	/**
	 * Perform scan.
	 * Invokes the scan of the java source in this project for taglib candidates.
	 *
   * @param  project  the project to scan
	 * @param  nodes    the set of nodes to scan
	 */
	private void scanNodesForTags(JBProject project, Node[] nodes) {
		for (int index = 0; index < nodes.length; index++) {
			if (nodes[index] instanceof JavaFileNode) {
				scanFileForTags(project, (JavaFileNode)nodes[index]);
			}
			else if (nodes[index] instanceof PackageNode &&
					nodes[index].getParent() == project) {
				// Scan into packages, but only to one level
				scanNodesForTags(project, 
            ((PackageNode)nodes[index]).getDisplayChildren());
			}
		}
	}

	/**
	 * Scans a file looking for javax.servlet.jsp.tagext.Tag implementers.
	 *
   * @param  project  the project to scan
	 * @param  fileNode  the node whose content is to be scanned
	 */
	private void scanFileForTags(JBProject project,	JavaFileNode fileNode) {
		JotPackages jpackages = project.getJotPackages();
		JotFile jfile = jpackages.getFile(fileNode.getUrl());
		JotClass[] jclasses = jfile.getClasses();
		for (int index = 0; index < jclasses.length; index++) {
			if (implementsInterface(jclasses[index], TAG_INTERFACES)) {
				scanClassForTags(jclasses[index]);
			}
			if (implementsInterface(jclasses[index], LISTENER_INTERFACES)) {
				_listeners.add(
            new TaglibDescriptorListener(jclasses[index].getName(), ""));
			}
			if (instanceOf(jclasses[index], 
          "javax.servlet.jsp.tagext.TagExtraInfo")) {
				_tagExtraInfos.add(jclasses[index].getName());
			}
			else if (instanceOf(jclasses[index],
					"javax.servlet.jsp.tagext.TagLibraryValidator")) {
				_validators.add(getValidator(jclasses[index]));
			}
		}
		jpackages.release(jfile);
	}

	/**
	 * Scan an individual class for components of a tag.
	 * These include its class name, whether it supports body content,
	 * and any attributes it might use.
	 *
	 * @param  jclass  the class to be scanned
	 */
	private void scanClassForTags(JotClass jclass) {
		int pos = jclass.getName().lastIndexOf(".");
		String name =
  			(pos == -1 ? jclass.getName() : jclass.getName().substring(pos + 1));
		String info = "";
		JotFile jfile = jclass.getFile();
		if (jfile instanceof JotSourceFile) {
			JotComment jcomment = ((JotSourceFile)jfile).
          getComment((JotClassSource)jclass, BEFORE);
			if (jcomment != null) {
				info = jcomment.getCommentText();
			}
		}
		TaglibDescriptorTag tag = new TaglibDescriptorTag(jclass.getName(), name,
		  	(implementsInterface(jclass, BODY_INTERFACES) ?
  			TaglibDescriptorTag.CONTENT_JSP : TaglibDescriptorTag.CONTENT_EMPTY),
	  		info);
    tag.setDynamicAttributes(implementsInterface(jclass, DYNAMIC_INTERFACES));
		// Load attributes - based on set* methods
		JotMethod[] jmethods = jclass.getMethods();
		nextMethod:
		for (int index = 0; index < jmethods.length; index++) {
			name = jmethods[index].getName();
			if (name.substring(0, 3).equals("set")) {
        if (Arrays.binarySearch(STD_METHODS, name) > -1) {
					// Ignore standard methods
					continue nextMethod;
				}
        JotClass[] types = jmethods[index].getParameterTypes();
        if (types.length != 1) {
          // Setters have only a single parameter
          continue nextMethod;
        }
				TaglibDescriptorAttribute tagAttr = new TaglibDescriptorAttribute(
		  			name.substring(3, 4).toLowerCase() + name.substring(4), 
            types[0].getName());
        if (jfile instanceof JotSourceFile) {
          JotComment jcomment = ((JotClassSource)jclass).
              getComment((JotMethodSource)jmethods[index], BEFORE);
          if (jcomment != null) {
            tagAttr.setDescription(jcomment.getCommentText());
          }
        }
        tag.getAttributes().add(tagAttr);
			}
			else if (name.equals("doStartTag")) {
				// Also check for body content in Tag
				if (jmethods[index] instanceof JotMethodSource) {
					String body =
						  ((JotMethodSource)jmethods[index]).getCodeBlock().getText();
					if (body.indexOf("EVAL_BODY_TAG") > -1 ||
							body.indexOf("EVAL_BODY_INCLUDE") > -1 ||
							body.indexOf("EVAL_BODY_BUFFERED") > -1) {
						tag.setBodyContent(TaglibDescriptorTag.CONTENT_JSP);
					}
				}
			}
      else if (name.equals("doTag")) {
        // Also check for body content in SimpleTag
        if (jmethods[index] instanceof JotMethodSource) {
          String body =
              ((JotMethodSource)jmethods[index]).getCodeBlock().getText();
          if (body.indexOf("getJspBody") > -1) {
            tag.setBodyContent(TaglibDescriptorTag.CONTENT_JSP);
          }
        }
      }
		}
		_tags.add(tag);
	}

	/**
	 * Scan an individual class for components of a validator.
	 * These include its comments and any attributes it might use.
	 *
	 * @param  jclass  the class to be scanned
	 * @return  the validator information
	 */
	private TaglibDescriptorValidator getValidator(JotClass jclass) {
		TaglibDescriptorValidator validator =
  			new TaglibDescriptorValidator(jclass.getName(), "");
		JotFile jfile = jclass.getFile();
		if (jfile instanceof JotSourceFile) {
			JotComment jcomment = ((JotSourceFile)jfile).
          getComment((JotClassSource)jclass, BEFORE);
			if (jcomment != null) {
				validator.setDescription(jcomment.getCommentText());
			}
		}
		// Load parameters - based on set* methods
		JotMethod[] jmethods = jclass.getMethods();
		for (int index = 0; index < jmethods.length; index++) {
			String name = jmethods[index].getName();
			if (name.substring(0, 3).equals("set")) {
        // Ignore standard methods
				if (!name.equals("setInitParameters")) {
          String descr = "";
          if (jfile instanceof JotSourceFile) {
            JotComment jcomment = ((JotClassSource)jclass).
                getComment((JotMethodSource)jmethods[index], BEFORE);
            if (jcomment != null) {
              descr = jcomment.getCommentText();
            }
          }
          validator.getParameters().add(new TaglibDescriptorParameter(
              name.substring(3, 4).toLowerCase() + name.substring(4), 
              "", descr));
        }
			}
		}
		return validator;
	}

	/**
	 * Does this class descend from the specified class.
	 *
	 * @param  jclass     the class to be tested
	 * @param  className  the name of the interface
	 * @return  true if the class or one of its ancestors is the one specified,
	 *          false otherwise
	 */
	private boolean instanceOf(JotClass jclass, String className) {
		while (jclass != null && !jclass.getName().equals("java.lang.Object")) {
			if (jclass.getType().getName().equals(className)) {
				return true;
			}
			jclass = jclass.getSuperclass().getJotClass();
		}
		return false;
	}

	/**
	 * Does this class implement the specified interface.
	 *
	 * @param  jclass          the class to be tested
	 * @param  interfaceNames  the list of interface names to check
	 * @return  true if the class or one of its ancestors implements one of the
	 *          interfaces, false otherwise
	 */
	private boolean implementsInterface(JotClass jclass, 
      String[] interfaceNames) {
		while (jclass != null && !jclass.getName().equals("java.lang.Object")) {
			JotType[] jtypes = jclass.getInterfaces();
			for (int index = 0; index < jtypes.length; index++) {
        if (Arrays.binarySearch(interfaceNames, jtypes[index].getName()) > -1) {
          return true;
        }
        for (int index2 = 0; index2 < interfaceNames.length; index2++) {
  				if (interfaceNames[index2].endsWith(jtypes[index].getName())) {
            return true;
          }
				}
			}
			jclass = jclass.getSuperclass().getJotClass();
		}
		return false;
	}
}
