package wood.keith.opentools.wizards.jsptags;

import java.util.ArrayList;

/**
 * Encapsulate details about a tag within a JSP tag library.
 *
 * @author   Keith Wood
 * @version  1.0  9 November 2000
 * @version  3.0  10 February 2004 - JSP 2.0
 */
public class TaglibDescriptorTag extends TaglibDescriptorElement {

	/** Body content option - this tag has no body. */
  public static final int CONTENT_EMPTY = 0;
	/** Body content option - this tag contains other JSP. */
	public static final int CONTENT_JSP = 1;
	/** Body content option - this tag contains non-JSP text. */
	public static final int CONTENT_TAG = 2;
  /** Body content option - this tag contains only template text. */
  public static final int CONTENT_SCRIPTLESS = 3;

	private static final String[] CONTENT = 
      {"empty", "JSP", "tagdependent", "scriptless"};

	// Internal variables
	private String _className;
	private String _name;
	private int _bodyContent;
	private String _tagExtraInfoClass = "";
	private String _example = "";
  private boolean _dynamicAttributes = false;
	private ArrayList _attributes = new ArrayList();
	private ArrayList _variables = new ArrayList();

	/**
	 * Build a new tag.
	 *
	 * @param  className    the name of the implementing Java class
	 * @param  name         the name of the tag in a JSP page
	 * @param  bodyContent  the type of content for the tag -
	 *                      see CONTENT_* constants
   * @param  description  a description of the tag
	 */
	public TaglibDescriptorTag(String className, String name, int bodyContent,
			String description) {
		super();
		setTagClassName(className);
		setName(name);
		setBodyContent(bodyContent);
		setDescription(description);
	}

	private void setTagClassName(String value) { _className = value; }

	public String getTagClassName() { return _className; }

	public void setName(String value) { _name = value; }

	public String getName() { return _name; }

	public void setBodyContent(int value) {
		if ((value < CONTENT_EMPTY) || (value > CONTENT_SCRIPTLESS)) {
			_bodyContent = CONTENT_EMPTY;
		}
		else {
			_bodyContent = value;
		}
	}

	public int getBodyContent() { return _bodyContent; }

	public String getTagExtraInfoClass() { return _tagExtraInfoClass; }

	public void setTagExtraInfoClass(String value) { _tagExtraInfoClass = value; }

	public String getExample() { return _example; }

	public void setExample(String value) { _example = value; }

  public boolean isDynamicAttributes() { return _dynamicAttributes; }

  public void setDynamicAttributes(boolean value) { 
    _dynamicAttributes = value; 
  }

	/**
	 * Provide access to the attributes for this tag.
	 *
	 * @return  a list of TaglibDescriptorAttribute objects
	 */
	public ArrayList getAttributes() { return _attributes; }

	/**
	 * Provide access to the variables for this tag.
	 *
	 * @return  a list of TaglibDescriptorVariable objects
	 */
	public ArrayList getVariables() { return _variables; }

	public String toString() { return getTagClassName(); }

	/**
	 * Format the tag for inclusion in a taglib descriptor file.
	 *
	 * @param  jspVersion  the version of JSP in use
	 * @return  the tag definition in XML
	 */
	public String toXml(String jspVersion) {
		StringBuffer buffer = new StringBuffer("\n  <tag>\n");
    if (jspVersion.equals(JSPTagProperties.JSP_20)) {
      addToXml20(buffer);
    }
     else if (jspVersion.equals(JSPTagProperties.JSP_12)) {
      addToXml12(buffer);
    }
    else {
      addToXml11(buffer);
    }
    buffer.append("  </tag>\n");
		return buffer.toString();
	}

  /**
   * Format the tag for inclusion in a JSP 2.0 taglib descriptor file.
   * 
   * @param  buffer  the string buffer to add to
   */
  private void addToXml20(StringBuffer buffer) {
    addDescriptionGroup(buffer, "    ");
    addXmlElement(buffer, "name", getName(), "    ");
    addXmlElement(buffer, "tag-class", getTagClassName(), "    ");
    addXmlElement(buffer, "tei-class", getTagExtraInfoClass(), "    ");
    addXmlElement(buffer, "body-content", CONTENT[getBodyContent()], "    ");
    for (int index = 0; index < _variables.size(); index++) {
      buffer.append(((TaglibDescriptorVariable)_variables.get(index)).
          toXml(JSPTagProperties.JSP_20));
    }
    for (int index = 0; index < _attributes.size(); index++) {
      buffer.append(((TaglibDescriptorAttribute)_attributes.get(index)).
          toXml(JSPTagProperties.JSP_20));
    }
    addXmlElement(buffer, "dynamic-attributes", 
        String.valueOf(isDynamicAttributes()), "    ");
    addXmlElement(buffer, "example", getExample(), "    ");
  }

  /**
   * Format the tag for inclusion in a JSP 1.2 taglib descriptor file.
   * 
   * @param  buffer  the string buffer to add to
   */
  private void addToXml12(StringBuffer buffer) {
    addXmlElement(buffer, "name", getName(), "    ");
    addXmlElement(buffer, "tag-class", getTagClassName(), "    ");
    addXmlElement(buffer, "tei-class", getTagExtraInfoClass(), "    ");
    addXmlElement(buffer, "body-content", CONTENT[getBodyContent()], "    ");
    addXmlElement(buffer, "display-name", getDisplayName(), "    ");
    addXmlElement(buffer, "small-icon", getSmallIcon(), "    ");
    addXmlElement(buffer, "large-icon", getLargeIcon(), "    ");
    addXmlElement(buffer, "description", getDescription(), "    ");
    for (int index = 0; index < _variables.size(); index++) {
      buffer.append(((TaglibDescriptorVariable)_variables.get(index)).
          toXml(JSPTagProperties.JSP_12));
    }
    for (int index = 0; index < _attributes.size(); index++) {
      buffer.append(((TaglibDescriptorAttribute)_attributes.get(index)).
          toXml(JSPTagProperties.JSP_12));
    }
    addXmlElement(buffer, "example", getExample(), "    ");
  }

  /**
   * Format the tag for inclusion in a JSP 1.1 taglib descriptor file.
   * 
   * @param  buffer  the string buffer to add to
   */
  private void addToXml11(StringBuffer buffer) {
    addXmlElement(buffer, "name", getName(), "    ");
    addXmlElement(buffer, "tagclass", getTagClassName(), "    ");
    addXmlElement(buffer, "teiclass", getTagExtraInfoClass(), "    ");
    addXmlElement(buffer, "bodycontent", CONTENT[getBodyContent()], "    ");
    addXmlElement(buffer, "info", getDescription(), "    ");
    for (int index = 0; index < _attributes.size(); index++) {
      buffer.append(((TaglibDescriptorAttribute)_attributes.get(index)).
          toXml(JSPTagProperties.JSP_11));
    }
  }
}
