package wood.keith.opentools.wizards.jsptags;

import java.io.File;
import java.util.ResourceBundle;
import javax.swing.ImageIcon;
import javax.swing.JOptionPane;

import com.borland.jbuilder.node.JBProject;
import com.borland.jbuilder.personality.WebAppPersonality;
import com.borland.primetime.PrimeTime;
import com.borland.primetime.help.HelpManager;
import com.borland.primetime.help.ZipHelpTopic;
import com.borland.primetime.ide.Browser;
import com.borland.primetime.node.FileNode;
import com.borland.primetime.personality.Personality;
import com.borland.primetime.util.VetoException;
import com.borland.primetime.vfs.Buffer;
import com.borland.primetime.vfs.Url;
import com.borland.primetime.wizard.BasicWizard;
import com.borland.primetime.wizard.Wizard;
import com.borland.primetime.wizard.WizardAction;
import com.borland.primetime.wizard.WizardHost;
import com.borland.primetime.wizard.WizardManager;
import com.borland.primetime.wizard.WizardPage;

/**
 * This class defines the "Taglib Descriptor" wizard at the highest level.
 * It sets the dialog pages to be displayed and generates the
 * TLD file as defined by user input.
 *
 * @author   Keith Wood
 * @version  1.0  9 November 2000
 * @version  2.0  6 February 2002 - added JSP 1.2 support
 * @version  2.1  16 June 2002
 * @version  3.0  10 February 2004 - JSP 2.0
 */
public class TaglibDescriptorWizard extends BasicWizard {

  private static final ResourceBundle res = Res.getResource();

  public static final String VERSION = "3.0";
  public static final String TITLE = res.getString("TLDWizardTitle");

  // Internal variables
  private TaglibDescriptorWizardPage1 _taglibPage1 = null;
  private TaglibDescriptorWizardPage2 _taglibPage2 = null;
  private TaglibDescriptorWizardPage3 _taglibPage3 = null;
  private TaglibDescriptorWizardPage4 _taglibPage4 = null;
  private TaglibDescriptorWizardPage5 _taglibPage5 = null;
  private TaglibDescriptorWizardPage6 _taglibPage6 = null;
  private String _jspVersion = JSPTagProperties.JSP_11;
  private TaglibDescriptorScanner _scanner;
  private ClassLoader _classLoader = new ProjectClassLoader();

  /**
   * Register the "Taglib Descriptor" tool.
   *
   * Provides the needed OpenTools interface required to register the
   * WizardAction which defines and creates this wizard.
   *
   * @param  majorVersion  the major version of the current OpenTools API
   * @param  minorVersion  the minor version of the current OpenTools API
   */
  public static void initOpenTool(byte majorVersion, byte minorVersion) {
    if (majorVersion != PrimeTime.CURRENT_MAJOR_VERSION) {
      return;
    }
    WizardManager.registerWizardAction(
        TaglibDescriptorWizard.WIZARD_TaglibDescriptor);
    if (PrimeTime.isVerbose()) {
      System.out.println("Loaded Taglib Descriptor wizard v" + VERSION);
      System.out.println("Written by Keith Wood (kbwood@iprimus.com.au)");
    }
  }

  /**
   * Definition of "Taglib Descriptor" WizardAction.
   *
   * This is an OpenTools registered action defining where the wizard will
   * appear in the IDE and a factory for creating each instance of this wizard.
   */
  public static final WizardAction WIZARD_TaglibDescriptor = new WizardAction(
      res.getString("TLDWizardName"), res.getString("TLDWizardChar").charAt(0),
      res.getString("TLDWizardTip"),
      new ImageIcon(ClassLoader.getSystemResource(
      "wood/keith/opentools/wizards/jsptags/taglib16.gif")),
      new ImageIcon(ClassLoader.getSystemResource(
      "wood/keith/opentools/wizards/jsptags/taglib32.gif")), true,
      res.getString("ObjectGalleryPage")) {

    // Override method so wizard is disabled if no active project.
    public void update(Object source) {
      Browser browser = Browser.findBrowser(source);
      setEnabled((browser != null) && (browser.getActiveProject() != null));
    }

    protected Wizard createWizard() {
      return new TaglibDescriptorWizard();
    }

    /**
     * JBuilder 10+ - only display for the Web app personality.
     *
     * @return  the list of personalities to which this action applies
     */
    public Personality[] getPersonalities() {
      return new Personality[] {WebAppPersonality.webAppPersonality};
    }
  };

  /**
   * Wizard startup.
   *
   * "Taglib Descriptor" wizard is to become visible,
   * create and order the wizard pages to be displayed.
   *
   * @param  host  the WizardHost that owns this wizard instance.
   * @return  the initial WizardPage to show.
   */
  public WizardPage invokeWizard(WizardHost host) {
    Browser browser = host.getBrowser();
    JBProject project = (JBProject)browser.getActiveProject();
    _jspVersion = JSPTagProperties.JSP_VERSION.getValue();
    // Find candidate classes
    _scanner = new TaglibDescriptorScanner(project);
    // Name flags presence of any suitable tags
    String name = null;
    if (_scanner.getTags().size() > 0) {
      // Make TLD internal name default to the same as project name
      name = project.getDisplayName();
      int pos = name.lastIndexOf(".");
      name = (pos == -1 ? name : name.substring(0, pos));
    }
    setWizardTitle(TITLE);
    _taglibPage1 =
        new TaglibDescriptorWizardPage1(this, name, getFileName(project));
    addWizardPage(_taglibPage1);
    if (_scanner.getTags().size() > 0) {
      // No subsequent pages if no tags found
      _taglibPage2 = new TaglibDescriptorWizardPage2(
          this, _scanner.getValidators());
      _taglibPage3 = new TaglibDescriptorWizardPage3(
          this, _scanner.getListeners());
      _taglibPage4 = new TaglibDescriptorWizardPage4(
          this, _scanner.getTagFiles());
      _taglibPage5 = new TaglibDescriptorWizardPage5(this);
      _taglibPage6 = new TaglibDescriptorWizardPage6(
          this, _scanner.getTags(), _scanner.getTagExtraInfos());
      setJSPVersion(_jspVersion);
      addWizardPage(_taglibPage6);
    }
    host.setFinishEnabled(true);
    return super.invokeWizard(host);
  }

  /**
   * Retrieve the version of JSP in use.
   *
   * @return  the JSP version
   */
  public String getJSPVersion() { return _jspVersion; }

  /**
   * Keep track of JSP version as selected by user on page 1 -
   * updating other pages as needed.
   *
   * @param  value  the version of JSP in use
   */
  public void setJSPVersion(String value) {
    if (_scanner.getTags().size() == 0) {
      return;
    }
    _jspVersion = value;
    if (!_jspVersion.equals(JSPTagProperties.JSP_11)) {
      if (!pages.contains(_taglibPage2)) {
        addWizardPage(1, _taglibPage2);
        addWizardPage(2, _taglibPage3);
      }
    }
    else {
      removeWizardPage(_taglibPage2);
      removeWizardPage(_taglibPage3);
    }
    if (_jspVersion.equals(JSPTagProperties.JSP_20)) {
      if (!pages.contains(_taglibPage4)) {
        addWizardPage(3, _taglibPage4);
        addWizardPage(4, _taglibPage5);
      }
    }
    else {
      removeWizardPage(_taglibPage4);
      removeWizardPage(_taglibPage5);
    }
    _taglibPage2.setJSPVersion(_jspVersion);
    _taglibPage3.setJSPVersion(_jspVersion);
    _taglibPage4.setJSPVersion(_jspVersion);
    _taglibPage5.setJSPVersion(_jspVersion);
    _taglibPage6.setJSPVersion(_jspVersion);
  }

  /**
   * Compute the suggested name for the TLD file.
   *
   * @param  project  the current project
   * @return  the TLD file name
   */
  private String getFileName(JBProject project) {
    String name = project.getDisplayName();
    name = (name.lastIndexOf('.') > -1 ?
        name.substring(0, name.lastIndexOf('.')) : name);
    return project.getUrl().getParent().
        getRelativeUrl(JSPTagProperties.TLD_LOCATION.getValue()).
        getRelativeUrl(name + ".tld").getFile();
  }

  /**
   * Supply the current project.
   *
   * @return  the current project
   */
  public JBProject getProject() {
    return (JBProject)wizardHost.getBrowser().getActiveProject();
  }

  /**
   * Supply a class loader that looks on the project path.
   *
   * @return  the project-based class loader
   */
  public ClassLoader getClassLoader() {
    return _classLoader;
  }

  /**
   * Creates/replaces file node.
   *
   * From the given parameters create/replace a project file node for the
   * "Taglib Descriptor" file to be written.
   *
   * @param  project      the JBProject in which the node is created.
   * @param  fileName     the name of the file to be created.
   * @return  the FileNode which was created.
   */
  protected static FileNode createNode(JBProject project,	String fileName) {
    Url filePath = new Url(new File(fileName));
    // getNode() will create a node if does not exist
    FileNode newNode = project.getNode(filePath);
    newNode.setParent(project);
    return newNode;
  }

  /**
   * Perform code generation.
   *
   * Produces the "Taglib Descriptor" file as defined by user input on the
   * wizard pages. It is added as a node to the currently active project
   * and replaces any previously generated file of the same name.
   *
   * @throws  VetoException
   */
  protected void finish() throws VetoException {
    if (_scanner.getTags().size() == 0) {
      return;
    }

    Browser browser = wizardHost.getBrowser();
    JBProject project = (JBProject)browser.getActiveProject();
    // Save selected JSP version for next time
    JSPTagProperties.JSP_VERSION.setValue(_jspVersion);
    // Save TLD location (directory relative to project) for next time
    JSPTagProperties.TLD_LOCATION.setValue(
        project.getUrl().getParent().getRelativePath(
        new Url(new File(_taglibPage1.getFileName())).getParent()));
    FileNode textNode = createNode(project, _taglibPage1.getFileName());
    try {
      Buffer buf = textNode.getBuffer();
      TaglibDescriptorGenerator generator = new TaglibDescriptorGenerator(
          _taglibPage1.getShortName(),
          _taglibPage1.getUri(), _taglibPage1.getDisplayName(),
          _taglibPage1.getSmallIconName(), _taglibPage1.getLargeIconName(),
          _taglibPage1.getDescription(), _taglibPage2.getValidator(),
          _taglibPage3.getListeners(), _taglibPage6.getTags(),
          _taglibPage4.getTagFiles(), _taglibPage5.getFunctions());
      buf.setContent(generator.toXml(_jspVersion).getBytes());
      // Cause the generated/updated file to be opened in the Content Pane.
      browser.setActiveNode(textNode, true);
    }
    catch (Exception ex) {
      ex.printStackTrace();
      JOptionPane.showMessageDialog(wizardHost.getBrowser(),
          (ex.getMessage() == null ? ex.getClass().getName() : ex.getMessage()),
          TITLE, JOptionPane.ERROR_MESSAGE);
      throw new VetoException();
    }
  }

  /**
   * Wizard termination.
   *
   * The Wizard has been completed (finished or cancelled) and any resource
   * cleanup to help the garbage collector can occur here.
   */
  public void wizardCompleted() {}

  /**
   * Provide help for the wizard.
   *
   * @param  page  the current page in the wizard
   * @param  host  the wizard host
   */
  public void help(WizardPage page, WizardHost host) {
    HelpManager.showHelp(new ZipHelpTopic(null, ClassLoader.getSystemResource(
        "wood/keith/opentools/wizards/jsptags/TaglibDescriptor" +
        ((TaglibDescriptorWizardPage)page).getPageNum() + ".html").toString()),
        page.getPageComponent(host));
  }
}
